import bs
import bsUtils
import bsSpaz
import random
import weakref
import bsUI
import bsInternal

gRandProfileIndex = 1
gLastWarnTime = 0

class PlayerReadyMessage(object):
    def __init__(self,chooser):
        self.chooser = chooser

class ChangeMessage(object):
    def __init__(self,what,value):
        self.what = what
        self.value = value

class Chooser(object):

    def __del__(self):
        # just kill off our base node; the rest should go down with it
        self._textNode.delete()
        
    def __init__(self,vPos,player,lobby):
        
        import bsInternal
        
        self._deekSound = bs.getSound('deek')
        self._clickSound = bs.getSound('click01')
        self._punchSound = bs.getSound('punch01')
        self._swishSound = bs.getSound('punchSwish')
        self._maskTexture = bs.getTexture('characterIconMask')
        self._vPos = vPos
        self._lobby = weakref.ref(lobby)
        self._player = player
        self._inited = False
        self._dead = False
        
        # note: this is just our local index of the team; *not* the team ID!
        self._selectedTeamIndex = self.getLobby().nextAddTeam

        # use this as an offset so input device 0 (the keyboard on desktops, touch-input on mobile) gets spaz as a default
        spazIndex = self.getLobby().characterNames.index('Spaz')

        # store a persistent random character index; we'll use this for the '_random' profile
        # lets use their inputDevice id to seed it.. this will give a persistent character
        # for them between games and will distribute characters nicely if everyone is random
        try: inputDeviceID = self._player.getInputDevice().getID()
        except Exception,e:
            print 'ERROR: exception getting inputDeviceID for lobby-chooser creation:',e
            inputDeviceID = 0
            import traceback
            traceback.print_stack()
            
        self._randomCharacterIndex = (inputDeviceID+spazIndex)%len(self.getLobby().characterNames)
        self._randomColor,self._randomHighlight = bsUtils.getPlayerProfileColors(None)

        # attempt to pick an initial profile based on what's been stored for this input device
        try:
            inputDevice = self._player.getInputDevice()
            name = inputDevice.getName()
            uniqueID = inputDevice.getUniqueIdentifier()
            self.playerProfileName = bs.getConfig()['Default Player Profiles'][name+' '+uniqueID]
            self.playerProfile = self.getLobby().playerProfileNames.index(self.playerProfileName)

        # well, that didn't work.. just start with the first profile (should be '_random')
        except Exception:
            global gRandProfileIndex
            # cycle through our non-random profiles once; after that, all is random
            if gRandProfileIndex < len(self.getLobby().playerProfileNames):
                self.playerProfile = gRandProfileIndex%len(self.getLobby().playerProfileNames)
            else:
                self.playerProfile = 0
            gRandProfileIndex += 1
            self.playerProfileName = self.getLobby().playerProfileNames[self.playerProfile]
            
        self.character = self._randomCharacterIndex
        self._color = self._randomColor
        self._highlight = self._randomHighlight
        self.ready = False
        self._textNode = bs.newNode('text',
                                    delegate=self,
                                    attrs={'position':(-100,self._vPos),
                                           'maxWidth':160,
                                           'shadow':0.5,
                                           'hAlign':'left',
                                           'vAlign':'center',
                                           'vAttach':'top'})
        # self._text = bs.NodeActor(bs.newNode('text',
        #                                      delegate=self,
        #                                      attrs={'position':(-100,self._vPos),
        #                                             'maxWidth':160,
        #                                             'shadow':0.5,
        #                                             'hAlign':'left',
        #                                             'vAlign':'center',
        #                                             'vAttach':'top'}))

        #bsUtils.animate(self._text.node,'scale',{0:0,100:1.0})
        bsUtils.animate(self._textNode,'scale',{0:0,100:1.0})
        self.icon = bs.newNode('image',
                               owner=self._textNode,
                               attrs={'position':(-130,self._vPos+20),
                                      'maskTexture':self._maskTexture,
                                      'attach':'topCenter'})

        bsUtils.animateArray(self.icon,'scale',2,{0:(0,0),100:(45,45)})

        self._setReady(False)

        # set our name to '<choosing player>' in case anyone asks..
        self._player.setName(bs.getResource('choosingPlayerText'))

        self.updateFromPlayerProfile()
        self.updatePosition()
        self._inited = True

    def getTeam(self):
        """ return the selected team """
        return self._lobby()._teams[self._selectedTeamIndex]()

    def getLobby(self):
        return self._lobby()

    def updateFromPlayerProfile(self):
        # set character based on profile; otherwise default to our pre-picked random
        try:
            # store the name even though we usually use index (in case the profile list changes)
            self.playerProfileName = self.getLobby().playerProfileNames[self.playerProfile]
            pc = self.getLobby().playerProfiles[self.playerProfileName]['character']
            self.character = self.getLobby().characterNames.index(pc)
            self._color,self._highlight = bsUtils.getPlayerProfileColors(self.playerProfileName)
        except Exception:
            self.character = self._randomCharacterIndex
            self._color = self._randomColor
            self._highlight = self._randomHighlight
        self._updateIcon()
        self._updateText()

    def reloadProfiles(self):
        # re-construct our profile index and other stuff since the profile list has changed
        try:
            self.playerProfile = self.getLobby().playerProfileNames.index(self.playerProfileName)
        except Exception:
            self.playerProfile = 0
            self.playerProfileName = self.getLobby().playerProfileNames[self.playerProfile]
        self.updateFromPlayerProfile()

    def updatePosition(self):
        # hmmm this shouldnt be happening..
        if not self._textNode.exists():
            'Err: chooser text nonexistant..'
            import traceback
            traceback.print_stack()
            return
        spacing = 350
        offs = spacing*-0.5*len(self.getLobby()._teams) + spacing*self._selectedTeamIndex + 250
        if len(self.getLobby()._teams) > 1: offs -= 35
        curPosition = self._textNode.position
        bsUtils.animateArray(self._textNode,'position',2,{0:self._textNode.position,100:(-100+offs,self._vPos+23)})
        bsUtils.animateArray(self.icon,'position',2,{0:self.icon.position,100:(-130+offs,self._vPos+22)})
        
    def getCharacterName(self):
        return self.getLobby().characterNames[self.character]
    
    def _doNothing(self):
        pass

    def _getName(self):
        name = self.getLobby().playerProfileNames[self.playerProfile]
        if name == '_random':
            try: inputDevice = self._player.getInputDevice()
            except Exception: inputDevice = None
            if inputDevice is not None:
                name = inputDevice._getDefaultPlayerName()
            else:
                name = 'Invalid'
        elif name == '_zNew':
            name = '<New Profile>'
        return name

    def _setReady(self,ready):

        import bsInternal
        
        profileName = self.getLobby().playerProfileNames[self.playerProfile]
        
        if not ready:
            self._player.assignInputCall('leftPress',bs.Call(self.handleMessage,ChangeMessage('team',-1)))
            self._player.assignInputCall('rightPress',bs.Call(self.handleMessage,ChangeMessage('team',1)))
            self._player.assignInputCall('bombPress',bs.Call(self.handleMessage,ChangeMessage('character',1)))
            self._player.assignInputCall('upPress',bs.Call(self.handleMessage,ChangeMessage('playerProfile',-1)))
            self._player.assignInputCall('downPress',bs.Call(self.handleMessage,ChangeMessage('playerProfile',1)))
            self._player.assignInputCall(('jumpPress','pickUpPress','punchPress'), bs.Call(self.handleMessage,ChangeMessage('ready',1)))
            self.ready = False
            self._updateText()
            self._player.setName('untitled')
        else:
            self._player.assignInputCall(('leftPress','rightPress', 'upPress','downPress',
                                          'jumpPress','bombPress','pickUpPress'),self._doNothing)
            self._player.assignInputCall(('jumpPress','bombPress','pickUpPress','punchPress'),
                                         bs.Call(self.handleMessage,ChangeMessage('ready',0)))

            # store the last profile picked by this input for reuse
            inputDevice = self._player.getInputDevice()
            name = inputDevice.getName()
            uniqueID = inputDevice.getUniqueIdentifier()
            try: deviceProfiles = bs.getConfig()['Default Player Profiles']
            except Exception: deviceProfiles = bs.getConfig()['Default Player Profiles'] = {}

            # make an exception if we have no profiles and are set to random
            # in that case we want to clear our setting so we pick up whatever
            # profile gets made
            if profileName == '_random' and len(self.getLobby().playerProfiles) == 1:
                try: del(deviceProfiles[name+' '+uniqueID])
                except Exception: pass
            else:
                deviceProfiles[name+' '+uniqueID] = profileName
            bs.writeConfig()

            # set this player's name
            self._player.setName(self._getName())
            self.ready = True
            self._updateText()

            # inform the session that this player is ready
            bs.getSession().handleMessage(PlayerReadyMessage(self))
                
    def handleMessage(self,m):

        if isinstance(m,ChangeMessage):

            # if we've been removed from the lobby, ignore this stuff
            if self._dead:
                print "WARNING: chooser got ChangeMessage after dying"
                return
            
            if not self._textNode.exists():
                bs.printError('got ChangeMessage after nodes died')
                return
            
            if m.what == 'team':
                if len(self.getLobby()._teams) > 1: bs.playSound(self._swishSound)
                self._selectedTeamIndex = (self._selectedTeamIndex + m.value) % len(self.getLobby()._teams)
                self._updateText()
                self.updatePosition()
                self._updateIcon()

            elif m.what == 'playerProfile':
                if len(self.getLobby().playerProfileNames) == 1:
                    global gLastWarnTime
                    if bs.getRealTime() - gLastWarnTime > 5000:
                        gLastWarnTime = bs.getRealTime()
                        bs.screenMessage(bs.getResource('noProfilesErrorText').replace('${NAME}',self._getName()))
                        bs.playSound(bs.getSound('error'))
                else:
                    # pick the next player profile and assign our name and character based on that
                    bs.playSound(self._deekSound)
                    self.playerProfile = (self.playerProfile + m.value) % len(self.getLobby().playerProfileNames)
                    self.updateFromPlayerProfile()
                
            elif m.what == 'character':
                bs.playSound(self._clickSound)
                self.character = (self.character + m.value) % len(self.getLobby().characterNames)
                self._updateText()
                self._updateIcon()

            elif m.what == 'ready':
                bs.playSound(self._punchSound)
                self._setReady(m.value)

    def _updateText(self):

        if self.ready:
            # once we're ready, we've saved the name, so lets ask the system for it
            # so we get appended numbers and stuff
            text = self._player.getName()
            text = text + ' ('+bs.getResource('readyText')+')'
        else:
            text = self._getName()

        canSwitchTeams = len(self.getLobby()._teams) > 1

        # flash as we're coming in
        finColor = bs.getSafeColor(self.getColor())+(1,)
        if not self._inited:
            bsUtils.animateArray(self._textNode,'color',4,{150:finColor,250:(2,2,2,1),350:finColor})
        else:
            # blend if we're in teams mode; switch instantly otherwise
            if canSwitchTeams:
                bsUtils.animateArray(self._textNode,'color',4,{0:self._textNode.color,100:finColor})
            else:
                self._textNode.color = finColor

        self._textNode.text = text

    def getColor(self):
        if self.getLobby()._useTeamColors:
            return self.getLobby()._teams[self._selectedTeamIndex]().color
        else:
            return self._color

    def getHighlight(self):
        # if we're using team colors we wanna make sure our highlight color
        # isn't too close to any other team's color
        highlight = list(self._highlight)
        if self.getLobby()._useTeamColors:
            for i,teamRef in enumerate(self.getLobby()._teams):
                team = teamRef()
                if i != self._selectedTeamIndex:
                    # find the dominant component this team's color
                    # and adjust ours so that component is not super-dominant
                    maxVal = 0
                    maxIndex = 0
                    for j in range(3):
                        if team.color[j] > maxVal:
                            maxVal = team.color[j]
                            maxIndex = j
                    thatColorForUs = highlight[maxIndex]
                    ourSecondBiggest = max(highlight[(maxIndex+1)%3],highlight[(maxIndex+2)%3])
                    diff = (thatColorForUs-ourSecondBiggest)
                    if diff > 0:
                        highlight[maxIndex] -= diff * 0.6
                        highlight[(maxIndex+1)%3] += diff * 0.3
                        highlight[(maxIndex+2)%3] += diff * 0.2
        return highlight

    def getPlayer(self):
        return self._player

    def _updateIcon(self):

        texName = bsSpaz.appearances[self.getLobby().characterNames[self.character]].iconTexture
        tintTexName = bsSpaz.appearances[self.getLobby().characterNames[self.character]].iconMaskTexture
        
        tex = bs.getTexture(texName)
        tintTex = bs.getTexture(tintTexName)

        self.icon.color = (1,1,1)
        self.icon.texture = tex
        self.icon.tintTexture = tintTex
        c = self.getColor()
        c2 = self.getHighlight()

        canSwitchTeams = len(self.getLobby()._teams) > 1

        # if we're initing, flash
        if not self._inited:
            bsUtils.animateArray(self.icon,'color',3,{150:(1,1,1),250:(2,2,2),350:(1,1,1)})

        # blend in teams mode; switch instantly in ffa-mode
        if canSwitchTeams:
            bsUtils.animateArray(self.icon,'tintColor',3,{0:self.icon.tintColor,100:c})
        else:
            self.icon.tintColor = c

        self.icon.tint2Color = c2

        finColor = (0.2+0.8*c[0],0.2+0.8*c[1],0.2+0.8*c[2])

        # store the icon info the the player
        self._player._setIconInfo(texName,tintTexName,c,c2)

class Lobby(object):

    def __del__(self):
        # reset any players that still have a chooser in us (should allow the choosers to die)
        players = [chooser._player for chooser in self.choosers if chooser._player.exists()]
        for p in players: p._reset()
        
    def __init__(self):
        
        session = bs.getSession()
        teams = session.teams if session._useTeams else None
        self._useTeamColors = session._useTeamColors

        if teams is not None:
            self._teams = [weakref.ref(team) for team in teams]
        else:
            self._dummyTeam = bs.Team()
            self._teams = [weakref.ref(self._dummyTeam)]
        
        vOffset = -150 if isinstance(session,bs.CoopSession) else -30
        
        #self._session = weakref.ref(session)
        self.choosers = []
        self.baseVOffset = vOffset
        self.updatePositions()
        self.nextAddTeam = 0
        #self._playerIcons = {}

        # grab available characters
        self.characterNames = bsSpaz.getAppearances()
        self.characterNames.sort()
        self.reloadProfiles()

        self._joinInfoText = None

    # def getSession(self):
    #     return self._session()
        
    def getChoosers(self):
        return self.choosers

    class JoinInfo(object):

        def __init__(self,lobby):
            self._text = bs.NodeActor(bs.newNode('text',
                                                 attrs={'position':(0,-40),
                                                        'hAttach':'center','vAttach':'top',
                                                        'hAlign':'center',
                                                        'color':(0.8,0.8,0.8,1.0),
                                                        'text':bs.getResource('waitingForLocalPlayersText')}))
            canSwitchTeams = (len(lobby._teams) > 1)
            self._state = 0

            pressToPunch = bs.getSpecialChar('leftButton')
            pressToPickup = bs.getSpecialChar('topButton')
            pressToBomb = bs.getSpecialChar('rightButton')

            # if we have a keyboard, grab keys for punch and pickup
            # FIXME - this of course is only correct locally; will need to change this for net games
            kb = bsInternal._getInputDevice('Keyboard','#1',exceptionOnNone=False)
            if kb is not None:
                punchKey = kb.getButtonName(bsUI.getControllerValue(kb,'buttonPunch'))
                pressToPunch = bs.getResource('orText').replace('${A}','\''+punchKey+'\'').replace('${B}',pressToPunch)
                bombKey = kb.getButtonName(bsUI.getControllerValue(kb,'buttonBomb'))
                pressToBomb = bs.getResource('orText').replace('${A}','\''+bombKey+'\'').replace('${B}',pressToBomb)
                joinStr = bs.getResource('pressPunchToJoinText')+'  < '+pressToPunch+' >'
            else:
                joinStr = bs.getResource('pressAnyButtonToJoinText')
                
            self._messages = ([joinStr]
                              + ([bs.getResource('pressToSelectTeamText').replace('${BUTTONS}',bs.getSpecialChar('leftArrow')+' '+bs.getSpecialChar('rightArrow'))] if canSwitchTeams else [])
                              + [bs.getResource('pressToSelectProfileText').replace('${BUTTONS}',bs.getSpecialChar('upArrow')+' '+bs.getSpecialChar('downArrow'))]
                              + [bs.getResource('pressToOverrideCharacterText').replace('${BUTTONS}',bs.getResource('bombBoldText'))+' < '+pressToBomb+' >'])
            
            self._timer = bs.Timer(4000,bs.WeakCall(self._update),repeat=True)

        def _update(self):
            self._text.node.text = self._messages[self._state]
            self._state = (self._state+1)%len(self._messages)

    def _createJoinInfo(self):
        return self.JoinInfo(self)

    def reloadProfiles(self):

        # grab available player profiles
        try: self.playerProfiles = dict(bs.getConfig()['Player Profiles'])
        except Exception: self.playerProfiles = {}

        # add in a random one so we're ok even if there's no user-created profiles
        self.playerProfiles['_random'] = {}
        self.playerProfileNames = self.playerProfiles.keys()
        self.playerProfileNames.sort(key=lambda x:x[0].lower())
        for chooser in self.choosers:
            try: chooser.reloadProfiles()
            except Exception:
                bs.printError('exception reloading profiles',printException=True)

    def updatePositions(self):
        self._vPos = -100+self.baseVOffset
        for chooser in self.choosers:
            chooser._vPos = self._vPos
            chooser.updatePosition()
            self._vPos -= 48

    def checkAllReady(self):
        for chooser in self.choosers:
            if not chooser.ready:
                return False
        return True

    def addChooser(self,player):
        self.choosers.append(Chooser(vPos=self._vPos,player=player,lobby=self))
        self.nextAddTeam = (self.nextAddTeam+1)%len(self._teams)
        self._vPos -= 48
        #bs.playSound(bs.getSound('dripity'))

    # def removeChooserPlayers(self):
    #     'kick players out of the game who currently have choosers'
    #     for c in self.choosers:
    #         try:
    #             player = c.getPlayer()
    #             player.removeFromGame()
    #         except Exception,e:
    #             print 'EXC in removeChooserPlayers:',e

    def removeAllChoosers(self):
        'called to remove all player choosers once they enter a game'
        self.choosers = []
        self.updatePositions()
        
    def removeChooser(self,player):
        'called to remove a player chooser once he enters a game'
        found = False
        for chooser in self.choosers:
            if chooser.getPlayer() is player:
                found = True
                # mark it as dead since there could be more
                # change-commands/etc coming in still for it;
                # want to avoid duplicate player-adds/etc
                chooser._dead = True
                self.choosers.remove(chooser)
                break
        if not found:
            print 'ERROR: removeChooser did not find player',player
            import traceback
            traceback.print_stack()
        elif chooser in self.choosers:
            print 'ERROR: removeChooser: chooser still present after removal',player
            import traceback
            traceback.print_stack()
            
        self.updatePositions()
